/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_ANONYMIZER_HXX_
# define WIFI_FRAME_FILTER_ANONYMIZER_HXX_

# include <cassert>
# include <stdexcept>

# include <boost/foreach.hpp>

# include <wipal/wifi/frame/filter/anonymizer.hh>

# include <wipal/wifi/frame/dissector/dissector.hh>

namespace wifi
{
  namespace frame
  {

    /*-----------------------.
    | Anonymization policies |
    `-----------------------*/

    namespace anon
    {

      inline
      addr
      dummy_mac_policy::operator [] (const addr& mac_addr) const
      {
	return mac_addr;
      }

      inline
      essid
      dummy_essid_policy::operator [] (const essid& essid) const
      {
	return essid;
      }

      template <class MACAnonPolicy, class ESSIDAnonPolicy>
      generic_policy<MACAnonPolicy, ESSIDAnonPolicy>::
      generic_policy(const MACAnonPolicy&	macp,
		     const ESSIDAnonPolicy&	essidp):
	MACAnonPolicy (macp),
	ESSIDAnonPolicy (essidp)
      {
      }

      template <class MACAnonPolicy, class ESSIDAnonPolicy>
      addr
      generic_policy<MACAnonPolicy, ESSIDAnonPolicy>::
      operator [] (const addr& mac_addr)
      {
	return MACAnonPolicy::operator [] (mac_addr);
      }

      template <class MACAnonPolicy, class ESSIDAnonPolicy>
      essid
      generic_policy<MACAnonPolicy, ESSIDAnonPolicy>::
      operator [] (const essid& essid)
      {
	return ESSIDAnonPolicy::operator [] (essid);
      }

    } // End of namespace wifi::frame::anon.


    namespace filter
    {

      /*--------------------.
      | anonymizer_iterator |
      `--------------------*/

      namespace internals
      {

	template <class I, class P,  class A, class B1, class B2>
	anonymizer_iterator<I, P, A, B1, B2>::
	anonymizer_iterator(const iterable_type& iterable, bool end):
	  A (iterable),
	  next_ (end? iterable.last_ : iterable.first_),
	  last_ (iterable.last_)
	{
	  if (not end)
	    increment();
	}

	template <class I, class P,  class A, class B1, class B2>
	bool
	anonymizer_iterator<I, P, A, B1, B2>::
	equal(const anonymizer_iterator& rhs) const
	{
	  if (not rhs.value())
	    return not this->value();

	  return next_ == rhs.next_;
	}

	template <class I, class P,  class A, class B1, class B2>
	void
	anonymizer_iterator<I, P, A, B1, B2>::increment()
	{
	  typedef typename I::value_type		input_frame;
	  typedef typename input_frame::pkthdr_ptr	pkthdr_ptr;
	  typedef typename input_frame::bytes_ptr	bytes_ptr;

	  if (next_ == last_)
	    {
	      this->value() = boost::none_t ();
	      return;
	    }

	  const input_frame&	  frm = *next_;
	  const pkthdr_ptr	  pcap = frm.pcap_header();
	  const uint8_t* const	  phy_u8 = frm.bytes().get();
	  const P* const	  phy = reinterpret_cast<const P*> (phy_u8);
	  const size_t		  caplen = pcap->caplen;
	  const bool		  swapped = frm.swapped();
	  const size_t		  phy_len = phy->len(caplen, swapped);

	  const uint8_t* const	  mac =
	    reinterpret_cast<const uint8_t*> (phy->decapsulate(caplen,
							       swapped));

	  const dissector<hooks>  d (mac, caplen - phy_len);
	  const hooks&		  h = d.hooks();

	  const size_t		  anon_caplen = (h.end?
						 h.end - mac + phy_len:
						 caplen);
	  pcapxx::pkthdr* const	  anon_pcap = new pcapxx::pkthdr (*pcap);
	  uint8_t* const	  anon_bytes = new uint8_t[anon_caplen];

	  memcpy(anon_bytes, phy, anon_caplen);
	  anonymize_addrs(phy_u8, h.addrs, anon_bytes, anon_caplen);
	  anon_pcap->caplen =
	    anonymize_ssid(frm, h.ssid, anon_bytes, anon_caplen);

	  this->value() = value_type (pkthdr_ptr (anon_pcap),
				      bytes_ptr (anon_bytes), frm.id());
	  ++next_;
	}

	template <class I, class P,  class A, class B1, class B2>
	void
	anonymizer_iterator<I, P, A, B1, B2>::
	anonymize_addrs(const uint8_t* const		phy,
			const std::set<const addr*>&	addrs,
			uint8_t* const			anon_bytes,
			const size_t			anon_caplen)
	{
	  BOOST_FOREACH(const addr* a, addrs)
	    {
	      const ptrdiff_t	offset = (reinterpret_cast<const uint8_t*>(a) -
					  phy);
	      addr* const	anon_addr = reinterpret_cast<addr*>
						(anon_bytes + offset);

	      if (offset + sizeof (*anon_addr) <= anon_caplen)
		*anon_addr = (*this)[*a];
	    }
	}

	template <class I, class P,  class A, class B1, class B2>
	size_t
	anonymizer_iterator<I, P, A, B1, B2>::
	anonymize_ssid(const typename I::value_type&	frm,
		       const ssid_elt* const		ssid,
		       uint8_t* const			anon_bytes,
		       const size_t			anon_caplen)
	{
	  if (not ssid)
	    return anon_caplen;
	  if (ssid->elt_id != 0)
	    {
	      std::cerr << frm << ": SSID element ID not found, skipping ("
			<< unsigned (ssid->elt_id) << " instead of 0)."
			<< std::endl;
	      return anon_caplen;
	    }

	  const size_t&		caplen = frm.pcap_header()->caplen;
	  const uint8_t* const	phy = frm.bytes().get();
	  const ptrdiff_t	ssid_offset = ssid->ssid - phy;

	  size_t		ssid_len = ssid->length;

	  if (ssid_offset > anon_caplen)
	    return anon_caplen;

	  if (ssid_len > 32)
	    {
	      std::cerr << frm << ": Invalid SSID length of "  << ssid_len
			<< ", truncating frame (should be between 0 and 32)."
			<< std::endl;
	      return ssid_offset;
	    }

	  if (ssid->ssid + ssid_len > phy + caplen)
	    {
	      std::cerr << frm << ": SSID overruns captured bytes, truncating."
			<< std::endl;
	      ssid_len = caplen - (ssid->ssid - phy);
	    }

	  const essid	ssid_str (*ssid, ssid_len);
	  essid		anon_ssid_str = (*this)[ssid_str];

	  if (anon_ssid_str.size() > ssid_len)
	    throw std::runtime_error ("Anonymized ESSID is too big");
	  anon_ssid_str.append(ssid_len - anon_ssid_str.size(), '\0');

	  if (ssid_offset + ssid_len > anon_caplen)
	    ssid_len = anon_caplen - ssid_offset;
	  memcpy(anon_bytes + ssid_offset, anon_ssid_str.data(), ssid_len);

	  return anon_caplen;
	}

	/*--------------------.
	| Anonymization hooks |
	`--------------------*/

	template <class I, class P, class A, class B1, class B2>
	anonymizer_iterator<I, P, A, B1, B2>::hooks::hooks():
	  ty (type::reserved),
	  end (0),
	  ssid (0)
	{
	}

	template <class I, class P, class A, class B1, class B2>
	void
	anonymizer_iterator<I, P, A, B1, B2>::
	hooks::addr_hook(const void* f, size_t, unsigned n, const addr& a)
	{
	  if (n == 1)
	    ty = type_of(f);

	  addrs.insert(&a);
	}

	template <class I, class P, class A, class B1, class B2>
	void
	anonymizer_iterator<I, P, A, B1, B2>::
	hooks::end_of_frame_headers_hook(const void* f,
					 size_t, const void* eofh)
	{
	  const uint8_t* const	eofh_u8 = static_cast<const uint8_t*> (eofh);

	  switch (ty)
	    {
	    case type::data:
	      end = eofh_u8;
	      break;

	    case type::management:
	      switch (subtype_of(f))
		{
		case mgt::subtype::assoc_req:
		  // 4 == sizeof (Caps. + Listen Interv.)
		  ssid = reinterpret_cast<const ssid_elt*> (eofh_u8 + 4);
		  break;

		case mgt::subtype::reassoc_req:
		  // 10 == sizeof (Caps. + Listen Interv. + Current AP addr.)
		  ssid = reinterpret_cast<const ssid_elt*> (eofh_u8 + 10);
		  break;

		case mgt::subtype::probe_req:
		  ssid = static_cast<const ssid_elt*> (eofh);
		  break;

		case mgt::subtype::probe_resp:
		case mgt::subtype::beacon:
		  // 12 == sizeof (Timestamp + Beacon Interv. + Caps.)
		  ssid = reinterpret_cast<const ssid_elt*> (eofh_u8 + 12);
		  break;
		}
	      break;
	    }
	}

      } // End of namespace wifi::frame::filter::internals.


      template <class I, class P, class A, class B>
      anonymizer<I, P, A, B>::anonymizer(const I& first, const I& last,
					 const A& anon_policy):
	A (anon_policy),
	first_ (first),
	last_ (last)
      {
      }

    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_FILTER_ANONYMIZER_HXX_
