/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_MERGE_HXX_
# define WIFI_FRAME_FILTER_MERGE_HXX_

extern "C"
{
# include <sys/time.h>
}

# include "merge.hh"

# include <wipal/wifi/frame/filter/time_adjuster.hh>

namespace wifi
{
  namespace frame
  {
    namespace filter
    {

      namespace internals
      {

	template <class HeaderType, class FD1>
	equal<HeaderType, FD1>::
	equal(const pcapxx::frame_descriptor<FD1>&	lhs,
	      const tool::endian::endianness&		phy_end):
	  lhs_ (lhs),
	  phy_end_ (phy_end)
	{
	}

	template <class HeaderType, class FD1>
	template <class FD2>
	bool
	equal<HeaderType, FD1>::
	operator == (const pcapxx::frame_descriptor<FD2>& rhs) const
	{
	  return HeaderType::eq_time_and_80211(lhs_, rhs, phy_end_);
	}

	template <class HT, class FD1>
	equal<HT, FD1>
	make_equal(const pcapxx::frame_descriptor<FD1>&	lhs,
		   const tool::endian::endianness&	phy_end)
	{
	  return equal<HT, FD1> (lhs, phy_end);
	}

	template <class HT, class FD1, class FD2>
	bool
	operator == (const pcapxx::frame_descriptor<FD2>&	lhs,
		     const equal<HT, FD1>&			rhs)
	{
	  return rhs == lhs;
	}

	template <class HT, template <class, class, class, class> class Merger,
		  class I, class F>
	struct provide_merge
	{
	  provide_merge(const I&			first2,
			const I&			last2,
			tool::opt::list&		options,
			F&				func):
	    first2_ (first2),
	    last2_ (last2),
	    options_ (options),
	    func_ (func)
	  {
	  }

	  template <class TimeAdjuster>
	  void
	  operator () (const TimeAdjuster& a)
	  {
	    typedef tool::endian::endianness	endianness;

	    const endianness& end = options_["endianness"].get<endianness>();

	    if (options_["ignore_noisy_prism"].get<bool>())
	      {
		typedef non_noisy_prism<I>			nnp;
		typedef typename nnp::const_iterator		n_iterator;
		typedef microseconds_stamper<n_iterator, HT>	us_stamper;
		typedef typename us_stamper::const_iterator	u_iterator;
		typedef typename TimeAdjuster::const_iterator	a_iterator;
		typedef
		  Merger<HT, a_iterator, u_iterator, tool::bottom> merger;

		nnp		n (first2_, last2_);
		us_stamper	s (n.begin(), n.end(), end);
		merger		m (a.begin(), a.end(),
				   s.begin(), s.end(), end);

		func_(m);
	      }
	    else
	      {
		typedef microseconds_stamper<I, HT>		us_stamper;
		typedef typename us_stamper::const_iterator	u_iterator;
		typedef typename TimeAdjuster::const_iterator	a_iterator;
		typedef
		  Merger<HT, a_iterator, u_iterator, tool::bottom> merger;

		us_stamper	s (first2_, last2_, end);
		merger		m (a.begin(), a.end(),
				   s.begin(), s.end(), end);

		func_(m);
	      }
	  }

	private:
	  const I&			first2_;
	  const I&			last2_;
	  tool::opt::list&		options_;
	  F&				func_;
	};

      } // End of namespace wifi::frame::filter::internals.

      template <class U, class HT,
		template <class, class, class> class Int,
		template <class, class, class, class> class Merger,
		class I1, class I2, class F>
      void
      provide_merge(const I1& first1, const I1&	last1,
		    const I2& first2, const I2&	last2,
		    tool::opt::list&	options,
		    F&			func)
      {
	internals::provide_merge<HT, Merger, I2, F> func2 (first2, last2,
							   options,
							   func);

	provide_time_adjuster<U, HT, Int>(first1, last1,
					  first2, last2,
					  options, func2);
      }

    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_FILTER_MERGE_HXX_
