/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <stdexcept>
#include <iostream>
#include <cstdlib>

#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include <wipal/tool/args.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/phy/empty.hh>
#include <wipal/phy/prism_header.hh>
#include <wipal/phy/rtap_header.hh>

namespace
{

  pcapxx::descriptor<>::link_type
  check_linktype(const pcapxx::list_of_traces<>& l,
		 const pcapxx::list_of_traces<>& r,
		 bool				 warn)
  {
    typedef pcapxx::descriptor<>::link_type	link_type;

#define P pcapxx::descriptor<>::PRISM_HEADER

    const link_type lt = l.empty() ? P : l.get(0).linktype();
    const link_type rt = r.empty() ? P : r.get(0).linktype();

#undef P

    l.expect(lt);
    r.expect(rt);

    if (warn and lt != rt)
      std::cerr << "WARNING: " << l.file_name() << " and " << r.file_name()
		<< " have different link types (" << lt << " and " << rt
		<< "), using " << lt << '.' << std::endl;

    return lt;
  }

  template <class D>
  bool
  eq_default(const pcapxx::frame_descriptor<D>& l,
	     const pcapxx::frame_descriptor<D>& r)
  {
    return l == r;
  }

  template <class D>
  bool
  eq_bytes(const pcapxx::frame_descriptor<D>& l,
	   const pcapxx::frame_descriptor<D>& r)
  {
    return
      0 == memcmp(l.bytes().get(), r.bytes().get(),
		  std::min(l.pcap_header()->caplen, r.pcap_header()->caplen));
  }

  template <class P, class D>
  bool
  eq_80211(const pcapxx::frame_descriptor<D>& l,
	   const pcapxx::frame_descriptor<D>& r)
  {
    return P::eq_80211(l, r);
  }

  template <class P, class D>
  bool
  eq_time_and_80211(const pcapxx::frame_descriptor<D>& l,
		    const pcapxx::frame_descriptor<D>& r)
  {
    return P::eq_time_and_80211(l, r);
  }

  template <class I1, class I2, class Eq>
  bool
  cmp(I1& l, const I1& lend,
      I2& r, const I2& rend,
      Eq eq)
  {
    while (l != lend and r != rend and eq(*l, *r))
      {
	++l;
	++r;
      }
    return l == lend and r == rend;
  }

  unsigned
  score_bytes(const void* const l, const void* const lend,
	      const void* const r, const void* const rend)
  {
    const uint8_t* const	lend_u8	= static_cast<const uint8_t*> (lend);
    const uint8_t* const	rend_u8	= static_cast<const uint8_t*> (rend);
    const uint8_t*		l_u8	= static_cast<const uint8_t*> (l);
    const uint8_t*		r_u8	= static_cast<const uint8_t*> (r);

    const size_t		l_size = lend_u8 - l_u8;
    const size_t		r_size = rend_u8 - r_u8;
    const size_t		size_diff (l_size < r_size?
					   r_size - l_size:
					   l_size - r_size);
    unsigned			s = 0;

    while (l_u8 != lend_u8 and r_u8 != rend_u8)
      if (*l_u8++ != *r_u8++)
	++s;
    return s + size_diff;
  }

  template <class I1, class I2>
  unsigned
  score_frames(const I1& l, const I1& lend,
	       const I2& r, const I2& rend)
  {
    assert(l != lend or r != rend);

    if (l == lend)
      return r->pcap_header()->caplen + sizeof (pcapxx::pkthdr);
    if (r == rend)
      return l->pcap_header()->caplen + sizeof (pcapxx::pkthdr);

    const uint8_t* const	l_bytes = l->bytes().get();
    const uint8_t* const	lend_bytes = l_bytes + l->pcap_header()->caplen;
    const uint8_t* const	r_bytes = r->bytes().get();
    const uint8_t* const	rend_bytes = r_bytes + r->pcap_header()->caplen;

    const pcapxx::pkthdr* const	l_pcap = l->pcap_header().get();
    const pcapxx::pkthdr* const lend_pcap = l_pcap + 1;
    const pcapxx::pkthdr* const	r_pcap = r->pcap_header().get();
    const pcapxx::pkthdr* const rend_pcap = r_pcap + 1;

    return (score_bytes(l_bytes, lend_bytes, r_bytes, rend_bytes) +
	    score_bytes(l_pcap,  lend_pcap,  r_pcap,  rend_pcap));
  }

  template <class I1, class I2, class Score>
  unsigned
  score(I1& l, const I1& lend,
	I2& r, const I2& rend,
	Score score)
  {
    I1		l_1st_diff = lend;
    I2		r_1st_diff = rend;
    bool	ldone = l == lend;
    bool	rdone = r == rend;
    unsigned	s = 0;

    while (not ldone or not rdone)
      {
	const unsigned t = score(l, lend, r, rend);

	if (t and l_1st_diff == lend)
	  {
	    assert(r_1st_diff == rend);
	    l_1st_diff = l;
	    r_1st_diff = r;
	  }
	s += t;
	if (not ldone)
	  ldone = ++l == lend;
	if (not rdone)
	  rdone = ++r == rend;
      }
    l = l_1st_diff;
    r = r_1st_diff;
    return s;
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      namespace opt = tool::opt;
      namespace args = tool::args;
      namespace def = tool::args::def;

      opt::list opts ((opt::name ("misc") -= opt::list
		       ((opt::name ("cmp") -= def::cmp())),

		       opt::name ("ui") -= opt::list
		       ((opt::name ("errors") -= def::ui_errors(),
			 opt::name ("summary") -= def::ui_summary()))));

      args::parse(argc, argv, opts, args::count (2), args::none);

      typedef pcapxx::list_of_traces<>		list_of_traces;
      typedef pcapxx::descriptor<>::link_type	link_type;
      typedef list_of_traces::const_iterator	iterator;
      typedef pcapxx::descriptor<>		descriptor;

      const args::input&	input = opt::get<args::input>(opts["proc"]
							      ["input"]);
      list_of_traces		lhs (input[0]);
      list_of_traces		rhs (input[1]);
      const link_type		type =
	check_linktype(lhs, rhs, opt::get<bool>(opts["ui"]["errors"]));

      iterator		l = lhs.begin();
      iterator		r = rhs.begin();
      const iterator	lend = lhs.end();
      const iterator	rend = rhs.end();

      bool		eq (false);

      switch (opt::get<args::comp_method>(opts["misc"]["cmp"]))
	{
	case args::cmp_default:
	  eq = cmp(l, lend, r, rend, eq_default<descriptor>);
	  break;

	case args::cmp_802:
	  switch (type)
	    {

#define WP_cmp_802_case(R, Unused, LT)					\
	    case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	      eq = cmp(l, lend, r, rend,				\
		       eq_80211<BOOST_PP_TUPLE_ELEM(2, 1, LT), descriptor>); \
	      break;

	      BOOST_PP_SEQ_FOR_EACH(WP_cmp_802_case, ~, WP_LINKTYPES)

#undef WP_cmp_802_case

	    default:
		args::unsupported_linktype(type);
	      break;
	    }
	  break;

	case args::cmp_bytes:
	  eq = cmp(l, lend, r, rend, eq_bytes<descriptor>);
	  break;

	case args::cmp_time:
	  switch (type)
	    {

#define WP_cmp_time_case(R, Unused, LT)					\
	    case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	      eq = cmp(l, lend, r, rend,				\
		       eq_time_and_80211<BOOST_PP_TUPLE_ELEM(2, 1, LT), \
					 descriptor>);			\
	      break;

	      BOOST_PP_SEQ_FOR_EACH(WP_cmp_time_case, ~, WP_LINKTYPES)

#undef WP_cmp_802_case

	    default:
		args::unsupported_linktype(type);
	      break;
	    }
	  break;

	case args::cmp_score:
	  {
	    const unsigned s = score(l, lend, r, rend,
				     score_frames<iterator, iterator>);
	    eq = 0 == s;
	    std::cout << s << " byte(s) differ." << std::endl;
	    break;
	  }

	default:
	  throw std::logic_error ("Unknown logic error.  Report to "
				  PACKAGE_BUGREPORT);
	}

      if (opt::get<bool>(opts["ui"]["summary"]))
	{
	  std::cout << argv[0] << ": "
		    << lhs.file_name() << " and " << rhs.file_name() << ' ';
	  switch (eq)
	    {
	    case true:
	      std::cout << "are equals (after endianness adjustments.)";
	      break;
	    case false:
	      std::cout << "differs, frame " << (l == lend ? *r: *l) << '.';
	      break;
	    }
	  std::cout << std::endl;
	}
      return eq ? EXIT_SUCCESS : EXIT_FAILURE;
    }
  catch (std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
