/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <stdexcept>
#include <iostream>

#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include <wipal/tool/args.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/wifi/frame/dissector/dissector.hh>
#include <wipal/wifi/frame/dissector/default_hooks.hh>
#include <wipal/wifi/frame/frame.hh>
#include <wipal/wifi/frame/mgt.hh>
#include <wipal/wifi/frame/ctl.hh>
#include <wipal/wifi/frame/filter/non_noisy_prism.hh>

namespace
{
  struct addr_hooks_base: public wifi::frame::dissector_default_hooks
  {
    addr_hooks_base(): addr_ (0)
    {
    }

    const wifi::addr*	addr_;
  };

  struct tx_hooks: public addr_hooks_base
  {
    void
    addr_hook(const void* f, size_t, unsigned i, const wifi::addr& a)
    {
      if (i != 2)
	return;

      const unsigned	type = wifi::frame::type_of(f);
      const unsigned	subtype = wifi::frame::subtype_of(f);

      if (type == wifi::frame::type::management or
	  type == wifi::frame::type::data or
	  (type == wifi::frame::type::control and
	   (subtype == wifi::frame::ctl::subtype::ps_poll or
	    subtype == wifi::frame::ctl::subtype::rts)))
	addr_ = &a;
    }
  };

  struct bssid_hooks: public addr_hooks_base
  {
    void
    addr_hook(const void* f, size_t, unsigned i, const wifi::addr& a)
    {
      switch (wifi::frame::type_of(f))
	{
	case wifi::frame::type::management:
	  if (i != 3)
	    return;
	  break;

	case wifi::frame::type::control:
	  switch (wifi::frame::subtype_of(f))
	    {
	    case wifi::frame::ctl::subtype::ps_poll:
	      if (i != 1)
		return;
	      break;

	    case wifi::frame::ctl::subtype::cf_end:
	    case wifi::frame::ctl::subtype::cf_end_cf_ack:
	      if (i != 2)
		return;
	      break;

	    default:
	      return;
	  }
	  break;

	case wifi::frame::type::data:
	  {
	    const unsigned dir = ((wifi::frame::to_ds(f)   ? 2 : 0) |
				  (wifi::frame::from_ds(f) ? 1 : 0));

	    if (2 < dir or i != 3 - dir)
	      return;
	  }
	  break;

	default: // Unkown frame type, implies unknown BSS.
	  return;
	}
      addr_ = &a;
    }
  };

  template <class PhyT, class Hooks>
  struct extractor
  {


    extractor(const wifi::addr&		addr,
	      const pcapxx::dumper&	dump): addr_ (addr),
					       dump_ (dump)
    {
    }

    template <class D>
    void
    operator () (const pcapxx::frame_descriptor<D>& f)
    {
      wifi::frame::dissector<Hooks> d = wifi::frame::dissect<PhyT, Hooks>(f);

      if (d.addr_ and *d.addr_ == addr_)
	dump_(f);
    }

  private:
    wifi::addr		addr_;
    pcapxx::dumper	dump_;
  };

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      namespace opt = tool::opt;
      namespace args = tool::args;

      opt::list opts ((opt::name ("internals") -= opt::list
		       ((opt::name ("ignore_noisy_prism") -=
					   args::def::ignore_noisy_prism()))));

      args::parse(argc, argv, opts, args::count (2), args::count (1));

      const args::input&	input = opt::get<args::input>(opts["proc"]
							      ["input"]);
      wifi::addr		a (input[0]);
      pcapxx::list_of_traces<>	t (input[1]);

      if (t.empty())
	return EXIT_SUCCESS;

      const args::output&	output = opt::get<args::output>(opts["proc"]
								["output"]);
      pcapxx::dumper		d = t.get(0).dumper(output[0]);

      if (opt::get<bool>(opts["internals"]["ignore_noisy_prism"]))
	t.expect(pcapxx::descriptor<>::PRISM_HEADER);

      switch (const pcapxx::descriptor<>::link_type lt = t.get(0).linktype())
	{
#define WP_case(R, Unused, LT)						\
	case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	  {								\
	    extractor<BOOST_PP_TUPLE_ELEM(2, 1, LT), WP_ADDR_HOOKS>	\
	      extract (a, d);						\
	    								\
	    if (opt::get<bool>(opts["internals"]["ignore_noisy_prism"])) \
	      {								\
		typedef pcapxx::list_of_traces<>::const_iterator	it_t; \
		typedef wifi::frame::filter::non_noisy_prism<it_t>	nnp_t; \
									\
		nnp_t (t.begin(), t.end()).for_each(extract);		\
	      }								\
	    else							\
	      t.for_each(extract);					\
	  }								\
	  break;

	  BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)

	default:
	  args::unsupported_linktype(lt);
	  break;
	}
    }
  catch (std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }

  return EXIT_SUCCESS;
}
