/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <iostream>
#include <stdexcept>
#include <boost/bind.hpp>
#include <boost/foreach.hpp>

#include <wipal/tool/args.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/wifi/frame/filter/microseconds_stamper.hh>
#include <wipal/wifi/frame/filter/non_noisy_prism.hh>

namespace
{

  std::ostream&
  operator << (std::ostream& os, struct timeval& tv)
  {
    return os << tool::microseconds (tv) << '\n';
  }

  void
  print_colheads(const tool::opt::list& opt, const std::string& filename)
  {
    const tool::opt::list&	ui = opt["ui"].get<tool::opt::list>();
    const bool&			indices = ui["indices"].get<bool>();
    const bool&			microseconds = ui["microseconds"].get<bool>();

    std::cout << filename << '\n';

    if (indices)
      std::cout << "Frame ID" << (microseconds? '\t' : '\n');
    if (microseconds)
      std::cout << "Microseconds\n";

    if (indices)
      std::cout << "========" << (microseconds? '\t' : '\n');
    if (microseconds)
      std::cout << "============\n";

    std::cout << std::flush;
  }

  template <class Frame>
  void
  list_frame(const Frame& f, const bool& indices, const bool& microseconds)
  {
    if (indices)
      std::cout << f.id() << (microseconds? '\t' : '\n');
    if (microseconds)
      std::cout << f.microseconds() << '\n';
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      namespace opt = tool::opt;
      namespace args = tool::args;
      namespace def = tool::args::def;

      typedef tool::endian::endianness		endianness;
      typedef pcapxx::list_of_traces<>		list_of_traces;

      opt::list opts ((opt::name ("internals") -= opt::list
		       (( opt::name ("endianness") -= def::endianness(),
			  opt::name ("ignore_noisy_prism") -=
						   def::ignore_noisy_prism())),
		       opt::name ("ui") -= opt::list
		       ((opt::name ("headers") -= def::ui_headers(),
			 opt::name ("indices") -= def::ui_indices(),
			 opt::name ("microseconds") -=
						    def::ui_microseconds()))));

      args::parse(argc, argv, opts, args::arbitrary, args::none);

      BOOST_FOREACH(list_of_traces t,
		    opt::get<args::input>(opts["proc"]["input"]))
	{
	  if (opt::get<bool>(opts["internals"]["ignore_noisy_prism"]))
	    t.expect(pcapxx::descriptor<>::PRISM_HEADER);

	  if (opt::get<bool>(opts["ui"]["headers"]))
	    print_colheads(opts, t.file_name());

	  const opt::list&	ui = opt::get<opt::list>(opts["ui"]);
	  const endianness&	end = opt::get<endianness>(opts["internals"]
							   ["endianness"]);

	  switch (const pcapxx::descriptor<>::link_type lt =
		    t.get(0).linktype())
	    {

#define WP_case(R, Unused, LT)						\
	    case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	      {								\
		using wifi::frame::filter::microseconds_stamper;	\
		using boost::bind;					\
									\
		typedef BOOST_PP_TUPLE_ELEM(2, 1, LT)	phy_header;	\
		typedef list_of_traces::const_iterator	iterator;	\
		typedef microseconds_stamper<iterator, phy_header>	\
							mss_stamper;	\
									\
		mss_stamper s (t.begin(), t.end(), end);		\
									\
		if (opt::get<bool>(opts["internals"]			\
				   ["ignore_noisy_prism"]))		\
		  {							\
		    using wifi::frame::filter::non_noisy_prism;		\
									\
		    typedef mss_stamper::const_iterator		it;	\
		    typedef non_noisy_prism<it>			nnp;	\
		    typedef nnp::const_iterator::value_type	frame;	\
									\
		    nnp	n (s.begin(), s.end());				\
									\
		    n.for_each(bind(&list_frame<frame>, _1,		\
				    opt::get<bool>(ui["indices"]),	\
				    opt::get<bool>(ui["microseconds"]))); \
		  }							\
		else							\
		  {							\
		    typedef						\
		      mss_stamper::const_iterator::value_type		\
		      frame;						\
									\
		    s.for_each(bind(&list_frame<frame>, _1,		\
				    opt::get<bool>(ui["indices"]),	\
				    opt::get<bool>(ui["microseconds"]))); \
		  }							\
	      }								\
	    break;

	      BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)

	    default:
	      args::unsupported_linktype(lt);
	      break;
	    }
	}
    }
  catch (std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
