/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <iostream>
#include <stdexcept>
#include <cstdlib>

#include <boost/format.hpp>
#include <boost/foreach.hpp>
#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include <wipal/tool/args.hh>
#include <wipal/tool/microseconds_stamp.hh>
#include <wipal/wifi/frame/filter/uniquely_identifiable.hh>
#include <wipal/wifi/addr_mapping.hh>
#include <wipal/pcap/list_of_traces.hh>

namespace
{

  template <class UniqueId>
  struct unique_frame_dumper
  {
    unique_frame_dumper(const tool::opt::list& options):
      microseconds_ (options["ui"]["microseconds"].get<bool>()),
      format_ ("%20s")
    {
    }

    void	operator () (const tool::microseconds_stamp<UniqueId>& frame)
    {
      std::cout << (format_ % frame);
      if (microseconds_)
	std::cout << '\t' << (format_ % frame.microseconds());
      std::cout << '\n';
    }

    void	print_colheads()
    {
      std::cout << (format_ % "Frame");
      if (microseconds_)
	std::cout << '\t' << (format_ % "Time (us)");
      std::cout << '\n'

		<< (format_ % "=====");
      if (microseconds_)
	std::cout << '\t' << (format_ % "=========");
      std::cout << std::endl;
    }

  private:
    const bool&		microseconds_;
    boost::format	format_;
  };

  struct main_helper
  {
    template <class UniqueId>
    int
    operator () (tool::opt::list& options) const
    {
      using tool::opt::get;
      using wifi::frame::filter::for_each_uniquely_identifiable;

      unique_frame_dumper<UniqueId>	dump (options);

      BOOST_FOREACH(pcapxx::list_of_traces<> traces,
		    get<tool::args::input>(options["proc"]["input"]))
	{
	  if (get<bool>(options["internals"]["ignore_noisy_prism"]))
	    traces.expect(pcapxx::descriptor<>::PRISM_HEADER);

	  if (get<bool>(options["ui"]["headers"]))
	    {
	      std::cout << traces.file_name() << ":\n";
	      dump.print_colheads();
	    }

	  if (not traces.empty())
	    switch (traces.get(0).linktype())
	      {

#define WP_case(R, Unused, LT)						\
	      case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
		traces.expect(pcapxx::descriptor<>::			\
			      BOOST_PP_TUPLE_ELEM(2, 0, LT));		\
	        for_each_uniquely_identifiable<UniqueId,		\
		  BOOST_PP_TUPLE_ELEM(2, 1, LT)>			\
		  (traces.begin(), traces.end(),			\
		   get<tool::opt::list>(options["internals"]), dump);	\
		break;

		BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)
#undef WP_case

	      default:
		  tool::args::unsupported_linktype(traces.get(0).linktype());
	      }
	}
      return EXIT_SUCCESS;
    }

  }; // End of struct main_helper.

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      namespace opt = tool::opt;
      namespace args = tool::args;
      namespace def = tool::args::def;

      opt::list	opts ((opt::name ("internals") -= opt::list
		       ((opt::name ("endianness") -= def::endianness(),
			 opt::name ("ignore_noisy_prism") -=
						     def::ignore_noisy_prism(),
			 opt::name ("mapping") -= def::mapping())),

		       opt::name ("misc") -= opt::list
		       ((opt::name ("mac_map") -= def::mac_map(),
			 opt::name ("attributes") -= def::attributes())),

		       opt::name ("ui") -= opt::list
		       ((opt::name ("headers") -= def::ui_headers(),
			 opt::name ("microseconds") -=
						    def::ui_microseconds()))));

      args::parse(argc, argv, opts, args::arbitrary, args::none);
      return args::dispatch_unique_id(opts, main_helper ());
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
